import mlflow
from mlflow.entities.run_status import RunStatus
import yaml
from datetime import datetime
import shutil
import os
import tempfile
from typing import Optional
import requests
import json
import uuid
import re
from shippedbrain import LOGIN_URL, UPLOAD_URL
import inspect

flavors = {
    "pyfunc": mlflow.pyfunc,
    "h2o": mlflow.h2o,
    "keras": mlflow.keras,
    "lightgbm": mlflow.lightgbm,
    "pytorch": mlflow.pytorch,
    "sklearn": mlflow.sklearn,
    "statsmodels": mlflow.statsmodels,
    "tensorflow": mlflow.tensorflow,
    "xgboost": mlflow.xgboost,
    "spacy": mlflow.spacy,
    "fastai": mlflow.fastai
}

# Key in MLmodel file
flavor_name = {
    "pyfunc": "python_function",
    "h2o": "h2o",
    "keras": "keras",
    "lightgbm": "lightgbm",
    "pytorch": "pytorch",
    "sklearn": "sklearn",
    "statsmodels": "statsmodels",
    "tensorflow": "tensorflow",
    "xgboost": "xgboost",
    "spacy": "spacy",
    "fastai": "fastai"
}

valid_model_name_regex = r"^[a-zA-Z]+[\w-]*$"

def _validate_model_name(model_name: str) -> bool:
    """ Validate model name - match with regex

    :param model_name: name of the model to publish on app.shippedbrain.com

    :return: True if name is valid, False otherwise
    """
    return type(model_name) is str and re.match(valid_model_name_regex, model_name) is not None


def _validate_run_id(mlflow_client: mlflow.tracking.MlflowClient,
                     run_id: str) -> bool:
    """ Validate user's run from existing model

    :param mlflow_client: MlflowClient instance
    :param run_id: run id of logged mlflow model

    :return: True if run id is valid, False otherwise
    """
    try:
        run_info = mlflow_client.get_run(run_id)

        if RunStatus.from_string(run_info.info.status) == mlflow.entities.run_status.RunStatus.FINISHED:
            return True

        return False
    except:
        raise Exception(f"Failed to get run with id {run_id}")


def _get_logged_model(run_id: str) -> dict:
    """ Get logged model from run id

    :param run_id: mlflow run id of the logged model

    :return: logged model py dict
    """
    _run = mlflow.get_run(run_id)
    # eval is needed, otherwise type is str
    log_model_history = eval(_run.data.to_dictionary()["tags"]["mlflow.log-model.history"])
    for logged_model in log_model_history:
        if run_id == logged_model['run_id']:
            return logged_model

    raise Exception(f"Could not find model with run_id={run_id}")


def _validate_model(run_id: str) -> bool:
    """ Validate logged mlflow model - look for input_example and signature

    :param run_id: the model's run id

    :return: True if model is valid, False otherwise
    """
    logged_model = _get_logged_model(run_id)
    # signature can be inferred from input_example
    try:
        _ = logged_model['signature']
    except Exception as e:
        raise Exception(
            f"Could not fetch signature for model with run id '{run_id}'. Please log a model with a valid "
            f"signature.\nLogged model {logged_model}. Error: {e}"
        )
    try:
        _ = logged_model['saved_input_example_info']
    except Exception as e:
        raise Exception(
            f"Could not fetch input example for model with run id '{run_id}'. Please log a model with a valid input "
            f"example.\nLogged model {logged_model}")

    return True


def _get_model_artifacts_path(run_id: str):
    try:
        logged_model = _get_logged_model(run_id)
        return logged_model['artifact_path']
    except:
        raise Exception(f"Could not find model artifact for run_id={run_id}")


def _download_artifacts(mlflow_client: mlflow.tracking.MlflowClient,
                        run_id: str,
                        tmpdirname: str) -> None:
    """Download artifacts from run

    :param mlflow_client: MlflowClient instance
    :param run_id: model's run id
    :param tmpdirname: a tempfile.TempDir path

    :return: None
    """
    model_artifacts = _get_model_artifacts_path(run_id)
    mlflow_client.download_artifacts(run_id=run_id, path=model_artifacts, dst_path=tmpdirname)


def _create_shipped_brain_yaml(model_name: str, model_artifacts_path: str, flavor: str, target_dir) -> str:
    """ Create a shipped-brain.yaml file in models artifacts

    :param model_name: the model's name on app.shippedbrain.com
    :param model_artifacts_path: the mlflow model's artifact path
    :param flavor: the flavor of the model (e.g. "pyfunc")
    :param target_dir: the directory to which to write the shipped-brain.yaml file

    :return: shipped-brain.yaml file absolute path
    """
    shipped_brain_yaml_file = os.path.join(target_dir, "shipped-brain.yaml")
    shipped_brain_yaml = {"model_name": model_name,
                          "model_artifacts_path": model_artifacts_path,
                          "flavor": flavor}
    with open(shipped_brain_yaml_file, "w") as yaml_file:
        yaml.dump(shipped_brain_yaml, yaml_file)

    return shipped_brain_yaml_file

def _update_MLmodel(new_run_id: str,
                    artifacts_uri: str,
                    model_artifacts_dir: str,
                    utc_time_created: Optional[datetime] = None) -> None:
    """ Update run_id field in MLmodel file

    :param new_run_id: new run id to update MLmodel
    :param artifacts_uri: artifacts path
    :param model_artifacts_dir: model artifacts dir in [artifacts_uri]
    :param utc_time_created: for tests only datetime to update in MLmodel

    :return: None
    """
    assert utc_time_created is None or type(utc_time_created) == datetime

    model_articats_path = os.path.join(artifacts_uri, model_artifacts_dir)
    mlmodel_path = os.path.join(model_articats_path, "MLmodel")

    with open(mlmodel_path, 'r+') as mlmodel_file:
        mlmodel_yaml = yaml.full_load(mlmodel_file)
        mlmodel_yaml['run_id'] = new_run_id
        mlmodel_yaml['utc_time_created'] = str(datetime.utcnow()) if not utc_time_created else str(utc_time_created)

    with open(mlmodel_path, 'w') as mlmodel_file:
        _ = yaml.dump(mlmodel_yaml, mlmodel_file)


def _zip_artifacts(model_artifacts_path: str) -> str:
    """ Zip artifacts from mlflow run

    :param model_artifacts_path: directory with model artifacts

    :return: temporary zip file name
    """
    file_name = str(uuid.uuid4())
    file_path = os.path.join(model_artifacts_path, file_name)
    shutil.make_archive(file_path, 'zip', model_artifacts_path)

    return file_path + ".zip"


def _unzip_artifacts(zipfile: str, target_dir: str) -> None:
    """ Unzip model artifacts file

    :param zipfile: absolute path to zip file
    :param target_dir: target directory to unpack file to

    :return: None
    """
    archive_format = "zip"

    file_name = zipfile.split("/")[-1]

    # Unpack the archive file
    shutil.unpack_archive(zipfile, target_dir, archive_format)

    # Delete zip file from target dir.
    try:
        os.remove(os.path.join(target_dir, file_name))
    except:
        print(f"Could not remove file {os.path.join(target_dir, file_name)} - target_dir='{target_dir}' file_name='{file_name}")

    print("Archive file unpacked successfully.")


def _login(email: str, password: str, login_url: str = LOGIN_URL) -> requests.Response:
    """ Login to Shipped Brain

    :param email: shipped brain account email
    :param password: shipped brain account password
    :param login_url: login url

    :return: requests.Response with auth token
    """

    headers = {
        'accept': 'application/json',
        'Content-Type': 'application/json',
    }

    data = json.dumps({"email": email, "password": password})

    response = requests.post(login_url, headers=headers, data=data)

    return response


def _upload_file(file_path: str,
                 auth_bearer: str,
                 upload_url: str = UPLOAD_URL) -> requests.Response:
    """Upload file request to Shipped Brain

    :param file_path: absolute path to zipfile to upload to app.shippedbrain.com
    :param auth_bearer: valid app.shippedbrain.com Authorization Bearer
    :param upload_url: file upload target url

    :return: upload response - request.Response
    """
    headers = {
        "Authorization": f"Bearer {auth_bearer}"
    }

    filename = file_path.split("/")[-1]

    with open(file_path, "rb") as file_obj:
        file = {
            "file": (filename, file_obj, "multipart/form-data"),
        }

        response = requests.post(upload_url, headers=headers, files=file)

    return response


def _is_valid_flavor(flavor: str) -> bool:
    """ Check if model flavor is valid

    :param flavor: the model flavor (e.g. "pyfunc")

    :return: True if valid flavor, False otherwise
    """
    return flavors.get(flavor) is not None


def _log_model(artifacts_path: str, model_artifacts_dir: str) -> mlflow.entities.Run:
    """ Log model from MLmodel file

    :param artifacts_path: absolute artifact dir. path
    :param model_artifacts_dir: relative model artifacts path in [tmpdir]

    :return: mlflow.entities.Run
    """

    with mlflow.start_run() as run:
        _update_MLmodel(run.info.run_id, artifacts_path, model_artifacts_dir)
        # USE mlflow.log_artifacts, otherwise root dir. is also copied
        mlflow.log_artifacts(artifacts_path)

    return run


def upload_model(flavor: str,
                 model_name: str,
                 input_example,
                 signature,
                 email: Optional[str] = None,
                 password: Optional[str] = None,
                 login_url: str = LOGIN_URL,
                 upload_url: str = UPLOAD_URL,
                 **kwargs) -> requests.Response:
    """ Publish trained model to shipped brain

    """
    assert _is_valid_flavor(flavor), f"Failed to validate model flavor. Bad model flavor '{flavor}'."

    email = email if email else os.getenv("SHIPPED_BRAIN_EMAIL")
    password = password if password else os.getenv("SHIPPED_BRAIN_PASSWORD")

    assert email, f"Bad email. Environment variable SHIPPED_BRAIN_EMAIL is not defined or you did not provide the email argument."
    assert password,f"Bad password. Environment variable SHIPPED_BRAIN_PASSWORD is not defined or you did not provide the password argument."
    assert _validate_model_name(model_name), f"Bad model name '{model_name}! Please provide a valid model name"

    model_run = _log_flavor(flavor=flavor, input_example=input_example, signature=signature, **kwargs)

    assert model_run is not None,\
        "An error occurred while trying to log model to user's mlflow tracking uri - model_run is None. "

    # No need to validate [model_run], [upload_run] function already does
    upload_response = upload_run(run_id=model_run.info.run_id,
                                 model_name=model_name, email=email,
                                 password=password,
                                 flavor=flavor,
                                 login_url=login_url,
                                 upload_url=upload_url)

    return upload_response


def upload_run(run_id: str,
               model_name: str,
               email: Optional[str] = None,
               password: Optional[str] = None,
               flavor: str = "pyfunc",
               login_url: str = LOGIN_URL,
               upload_url: str = UPLOAD_URL) -> requests.Response:
    """ Publish model to Shipped Brain from a logged mlflow model's run id

    :param run_id: run id of mlflow logged model
    :param model_name: name of the model to publish on app.shippedbrain.com
    :param email: shipped brain account email
    :param password: shipped brain account password
    :param flavor: flavor of the logged model
    :param login_url: login url
    :param upload_url: upload_url

    :return: model upload requests.Response from app.shippedbrain.com
    """

    client = mlflow.tracking.MlflowClient()

    email = email if email else os.getenv("SHIPPED_BRAIN_EMAIL")
    password = password if password else os.getenv("SHIPPED_BRAIN_PASSWORD")

    assert email, f"Bad email. Environment variable SHIPPED_BRAIN_EMAIL is not defined or you did not provide the email argument."
    assert password,f"Bad password. Environment variable SHIPPED_BRAIN_PASSWORD is not defined or you did not provide the password argument."
    assert _validate_model_name(model_name), f"Bad model name '{model_name}! Please provide a valid model name"
    assert _validate_run_id(client, run_id), f"The run id '{run_id}' you provided is not valid!"
    assert _validate_model(run_id), f"The model with '{run_id}' is not valid!"

    model_artifacts_path = _get_model_artifacts_path(run_id)

    with tempfile.TemporaryDirectory() as tmpdir:
        _download_artifacts(client, run_id, tmpdir)
        _create_shipped_brain_yaml(model_name, model_artifacts_path, flavor_name[flavor], tmpdir)
        zipped_file = _zip_artifacts(tmpdir)

        login_response = _login(email, password, login_url)
        assert login_response.status_code == 200, \
            f"Failed to login with status code {login_response.status_code}. Error: {login_response.text}"
        response_data = eval(login_response.text)

        auth_bearer = response_data["data"]["results"]["access_token"]

        upload_response = _upload_file(zipped_file, auth_bearer, upload_url)

        return upload_response

def _get_required_log_model_args(log_model_func: callable) -> list:
    """ Return function arguments with no default values (a.k.a. required)

    :param log_model_func: a python callable - expected mlflow.<flavor>.log_model

    :return: list of function arguments with no default value
    """
    signature = inspect.signature(log_model_func)
    return [k for k, v in signature.parameters.items() if v.default is inspect.Parameter.empty]


def _log_flavor(flavor: str,
                input_example,
                signature: mlflow.models.signature.ModelSignature,
                **kwargs) -> mlflow.entities.Run:
    # conda_env: str,
    # input_example: Union[pandas.core.frame.DataFrame, numpy.ndarray, dict, list],
    # signature: mlflow.models.signature.ModelSignature):
    """ Log model flavor to Shipped Brain - similar to mlflow.<flavour>.log_model

    :param model: path to serialized model
    :param flavor: a valid model flavor
    :param conda_env: path to conda environment file
    :param input_example: path to input example file
    :param signature: an mlflow.Signature

    :return: logged model mlflow.entities.Run instance
    """
    #print(f"[DEBUG] flavor={flavor}")
    #print(f"[DEBUG] kwargs={kwargs}")

    log_model_function_name = "log_model"

    # Required args by Shipped Brain
    # artifact_path arg. is required by [log_model] method
    # TODO handle model name

    assert _is_valid_flavor(flavor), f"Failed to validate model flavor. Bad model flavor '{flavor}'."
    assert hasattr(mlflow, flavor), f"Failed to log model. Could not find flavor '{flavor}' in mlflow." 

    flavor_module = eval(f"mlflow.{flavor}" )
    assert callable(getattr(flavor_module, log_model_function_name)), f"Failed to log model. Flavor '{flavor}' does not have {log_model_function_name} method."

    log_model_function = eval(f"mlflow.{flavor}.{log_model_function_name}")

    # Get log_model function required args
    log_function_required_args_set = set(_get_required_log_model_args(log_model_function))
    # remove kwargs
    log_function_required_args_set -= set(["kwargs"])
    
    user_kwargs_set = set(kwargs.keys())
    #print("[DEBUG] All required args", log_function_required_args_set)
    
    missing_args = log_function_required_args_set - user_kwargs_set

    #print("[DEBUG] MISSING ARGS:", missing_args)
    assert len(missing_args) == 0, f"Failed to log model. Missing arguments {missing_args}"

    # Log the model
    active_run = mlflow.active_run()
    if not active_run:
        active_run = mlflow.start_run()
        log_model_function(signature=signature, input_example=input_example, **kwargs)
        mlflow.end_run()
    else:
        log_model_function(signature=signature, input_example=input_example, **kwargs)

    return active_run


import os

MODEL_NAME = "Elastic-Net-Test"
MLRUNS_PATH = "/tmp/shippedbrain-test/mlruns"

LOGIN_URL_HTTP_BIN = "http://httpbin.org/post"
UPLOAD_URL_HTTP_BIN = "http://httpbin.org/post"

LOGIN_URL_HTTP = os.getenv("SHIPPED_BRAIN_LOGIN_URL")
UPLOAD_URL_HTTP = os.getenv("SHIPPED_BRAIN_UPLOAD_URL")


import pytest
import shutil
import os
import tests.resources.train as train
from tests import MLRUNS_PATH

def pytest_sessionstart(session):
    print('\n[STARTUP] Setting mlflow tracking uri')
    try:
        MLFLOW_TRACKING_URI  = f"file://{MLRUNS_PATH}"
        os.environ.setdefault("MLFLOW_TRACKING_URI", MLFLOW_TRACKING_URI)
        run_id = train.main()
        os.environ.setdefault("SHIPPED_BRAIN_TEST_RUN_ID", run_id.info.run_id)
    except Exception as e:
        print(f"[EXCEPTION] Failed to run startup routine! Error: {e}")

def pytest_sessionfinish(session, exitstatus):
    print("\n[TEARDOWN] Deleting mlruns. Test run_id:", os.environ.get("SHIPPED_BRAIN_TEST_RUN_ID"))
    #print(exitstatus)
    shutil.rmtree(MLRUNS_PATH)



import json
import pytest
import yaml
import tempfile
from shippedbrain import shippedbrain
import os
import mlflow
from tests import MODEL_NAME, MLRUNS_PATH, LOGIN_URL_HTTP_BIN, UPLOAD_URL_HTTP_BIN, LOGIN_URL_HTTP, UPLOAD_URL_HTTP
import tests.resources.train as train

client = mlflow.tracking.MlflowClient()

MLFLOW_TRACKING_URI = os.getenv("MLFLOW_TRACKING_URI")
RUN_ID = os.environ.get("SHIPPED_BRAIN_TEST_RUN_ID")

MODEL_ARTIFACTS_DIR = "model"
RUN_PATH = os.path.join(MLRUNS_PATH, f"0/{RUN_ID}/")


class TestShippedBrain:

    def test__validate_model_name(self):
        # positive
        assert shippedbrain._validate_model_name("Model")
        assert shippedbrain._validate_model_name("Model-")
        assert shippedbrain._validate_model_name("Mod_el")
        assert shippedbrain._validate_model_name("model2-a")
        # negative
        assert shippedbrain._validate_model_name("Mod/el") is False
        assert shippedbrain._validate_model_name("-Model") is False
        assert shippedbrain._validate_model_name("12model2-a") is False

    def test__is_valid_flavor(self):
        flavors = [
            "pyfunc",
            "h2o",
            "keras",
            "lightgbm",
            "pytorch",
            "sklearn",
            "statsmodels",
            "tensorflow",
            "xgboost",
            "spacy",
            "fastai"
        ]
        assert all([shippedbrain._is_valid_flavor(f) for f in flavors])

    def test__validate_run_id(self,):
        assert shippedbrain._validate_run_id(mlflow_client=client, run_id=RUN_ID), 'Failed to validate run_id.'

    def test__get_logged_model(self):
        logged_model = shippedbrain._get_logged_model(RUN_ID)
        print("LOGGED MODEL:", logged_model)

    def test__validate_model(self):
        assert shippedbrain._validate_run_id(client, RUN_ID), "Failed to validate model."

    def test__get_model_artifacts_path(self):
        model_artifacts_dir = "model"
        model_artifacts_dir_result = shippedbrain._get_model_artifacts_path(RUN_ID)
        assert model_artifacts_dir_result == model_artifacts_dir,\
            f"Failed to get model artifacts path. Return path is '{model_artifacts_dir_result}'"

    def test__download_artifacts(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            shippedbrain._download_artifacts(client, RUN_ID, tmpdir)
            # TODO compare dirs

    def test__create_shipped_brain_yaml(self):
        shipped_brain_yaml = {
            "model_name": MODEL_NAME,
            "flavor": shippedbrain.flavor_name["pyfunc"],
            "model_artifacts_path": MODEL_ARTIFACTS_DIR
        }
        with tempfile.TemporaryDirectory() as tmpdir:
            shipped_brain_yaml_path = os.path.join(tmpdir, "shipped-brain.yaml")

            shipped_brain_yaml_path_result = shippedbrain._create_shipped_brain_yaml(MODEL_NAME,
                                                                                     MODEL_ARTIFACTS_DIR,
                                                                                     shippedbrain.flavor_name["pyfunc"],
                                                                                     tmpdir)

            assert shipped_brain_yaml_path_result == shipped_brain_yaml_path

            with open(shipped_brain_yaml_path, "r") as file:
                shipped_brain_yaml_result = yaml.full_load(file)

                assert shipped_brain_yaml_result == shipped_brain_yaml, \
                    f"Bad shipped-brain.yaml file content! Content: {shipped_brain_yaml_result}"

    def test__update_MLmodel(self):
        import shutil
        from datetime import datetime

        mlmodel_path = './tests/resources/data/MLmodel'
        time_now = datetime.utcnow()
        yaml_base = dict()
        with open(mlmodel_path, "r") as file:
            yaml_base = yaml.full_load(file)

        with tempfile.TemporaryDirectory() as tmpdir:
            shutil.copy(mlmodel_path, tmpdir)
            shippedbrain._update_MLmodel(RUN_ID, tmpdir, "", utc_time_created=time_now)
            created_yaml_path = os.path.join(tmpdir, "MLmodel")

            with open(created_yaml_path, "r") as file:
                yaml_created = yaml.full_load(file)

                assert yaml_created.pop("run_id") == RUN_ID
                assert yaml_created.pop("utc_time_created") == str(time_now)
                yaml_base.pop("run_id")
                yaml_base.pop("utc_time_created")
                assert yaml_created == yaml_base

    def test__zip_artifacts(self):
        zip_file_path = shippedbrain._zip_artifacts(RUN_PATH)
        assert os.path.isfile(zip_file_path)

    def test__unzip_artifacts(self):
        zipfile_path = "./tests/resources/data/zipfile_example.zip"
        with tempfile.TemporaryDirectory() as tmpdir:
            shippedbrain._unzip_artifacts(zipfile_path, tmpdir)
            unzippedfile_path = os.path.join(tmpdir, "MLmodel")

            assert os.path.isfile(unzippedfile_path)
            # TODO read content ...

    # INTEGRATION
    def test__upload_file(self):
        import json
        zipfile_path = "./tests/resources/data/zipfile_example.zip"
        bearer = "some-auth-bearer"
        # Uses httpbin by default
        response = shippedbrain._upload_file(zipfile_path, bearer, UPLOAD_URL_HTTP_BIN)

        assert  response.status_code == 200
        response_json = json.loads(response.text)
        assert response_json["headers"]["Authorization"] == f"Bearer {bearer}"

    # INTEGRATION
    def test__login(self):
        email = "shippedbrain@mail.com"
        password = "passowrd"

        response = shippedbrain._login(email, password, LOGIN_URL_HTTP_BIN)
        response_text_as_dict = eval(response.text)
        response_data = eval(response_text_as_dict["data"])

        assert response.status_code == 200
        assert response_data["email"] == email
        assert response_data["password"] == password

    def test__log_model(self):
        artifacts_base = client.list_artifacts(RUN_ID)

        artifacts_path = os.path.join(RUN_PATH, "artifacts")
        logged_run = shippedbrain._log_model(artifacts_path, MODEL_ARTIFACTS_DIR)
        assert shippedbrain._validate_run_id(client, logged_run.info.run_id)
        artifacts_result = client.list_artifacts(logged_run.info.run_id)
        assert artifacts_base == artifacts_result

    def test__log_flavor_inside_run_context(self):
        log_model_option = {"flavor": "_log_flavor"}

        run = train.main(log_model_option=log_model_option, run_inside_mlflow_context=True)

        assert shippedbrain._validate_run_id(client, run.info.run_id), "Run id is not valid"
        assert shippedbrain._validate_model(run.info.run_id), "Logged model is not valid"

    def test__log_flavor_outside_run_context(self):
        log_model_option = {"flavor": "_log_flavor"}

        run = train.main(log_model_option=log_model_option, run_inside_mlflow_context=False)

        assert shippedbrain._validate_run_id(client, run.info.run_id), "Run id is not valid"
        assert shippedbrain._validate_model(run.info.run_id), "Logged model is not valid"

    def test_upload_run(self):
        response = shippedbrain.upload_run(run_id=RUN_ID,
                                           model_name="Test-Model-upload_run",
                                           email="bernardo@shippedbrain.com",
                                           password="password",
                                           login_url=LOGIN_URL_HTTP,
                                           upload_url=UPLOAD_URL_HTTP)

        assert response.status_code == 200

    def test_upload_model(self):
        train_x, train_y, test_x, test_y, alpha, l1_ratio = train.build_train()

        lr, predicted_qualities, signature, rmse, mae, r2 = train.train_and_eval(alpha,
                                                                                  l1_ratio,
                                                                                  train_x,
                                                                                  train_y,
                                                                                  test_x,
                                                                                  test_y)
        input_example = test_x.iloc[0:2]

        MAX_RETRIES = 3
        curr_try = 0
        import  time
        while curr_try < MAX_RETRIES:
            curr_try += 1
            response = shippedbrain.upload_model(flavor="sklearn",
                                                 model_name="Test-Model-upload_model",
                                                 email="bernardo@shippedbrain.com",
                                                 password="password",
                                                 signature=signature,
                                                 input_example=input_example,
                                                 sk_model=lr,
                                                 artifact_path="model",
                                                 login_url=LOGIN_URL_HTTP,
                                                 upload_url=UPLOAD_URL_HTTP
                                                 )
            # FORBIDDEN, server upload concurrency restrictions
            if response.status_code == 403:
                time.sleep(2 ** curr_try)
            else:
                break
        assert response.status_code == 200




import click
import os
import shippedbrain

@click.group()
@click.version_option(version="0.1", help="Get current installed shippedbrain version in system")
@click.help_option(help="Get help on how to use shippedbrain")
def cli():
    """Shipped Brain cli - Upload your models to app.shippedbrain.com"""
    pass

@cli.command("upload")
@click.option("--run_id", "-r", help="The run_id of logged mlflow model", type=str, required=True)
@click.option("--model_name", "-m", help="The model name to display in app.shippedbrain.com", type=str, required=True)
@click.option("--flavor", "-f", help="The mlflow flow flavor of the model", default="pyfunc", type=str)
@click.help_option(help="Get help on how to use the 'upload' command")
def upload(run_id: str, model_name: str, flavor: str):
    """Deploy a model to app.shippedbrain.com - create a REST endpoint and get a hosted model web page

    Set the environment variable SHIPPED_BRAIN_EMAIL to skip de email prompt
    """

    email = os.getenv("SHIPPED_BRAIN_EMAIL")
    password = os.getenv("SHIPPED_BRAIN_PASSWORD")

    if email:
        click.echo("Using environment variable SHIPPED_BRAIN_EMAIL")
        click.echo(f"email: {email}")
    else:
        email = click.prompt(text="email")

    if password:
        click.echo("Using environment variable SHIPPED_BRAIN_PASSWORD")
    else:
        password = click.prompt(text="password", hide_input=True)

    shippedbrain.upload(email=email, password=password, run_id=run_id, model_name=model_name)

if __name__ == "__main__":
    cli()


# The data set used in this example is from http://archive.ics.uci.edu/ml/datasets/Wine+Quality
# P. Cortez, A. Cerdeira, F. Almeida, T. Matos and J. Reis.
# Modeling wine preferences by data mining from physicochemical properties. In Decision Support Systems, Elsevier, 47(4):547-553, 2009.

"""
CREATE LOCAL mlruns
"""

import warnings
import sys

import pandas as pd
import numpy as np
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score
from sklearn.model_selection import train_test_split
from sklearn.linear_model import ElasticNet
from urllib.parse import urlparse
import mlflow.sklearn
from mlflow.models.signature import infer_signature

from  tests import MODEL_NAME

import logging

from shippedbrain import shippedbrain

logging.basicConfig(level=logging.WARN)
logger = logging.getLogger(__name__)


def eval_metrics(actual, pred):
    rmse = np.sqrt(mean_squared_error(actual, pred))
    mae = mean_absolute_error(actual, pred)
    r2 = r2_score(actual, pred)
    return rmse, mae, r2

def train_and_eval(alpha, l1_ratio, train_x, train_y, test_x, test_y):
    lr = ElasticNet(alpha=alpha, l1_ratio=l1_ratio, random_state=46)

    lr.fit(train_x, train_y)

    predicted_qualities = lr.predict(test_x)

    (rmse, mae, r2) = eval_metrics(test_y, predicted_qualities)

    # Infer model signature
    signature = infer_signature(test_x, predicted_qualities)

    return lr, predicted_qualities, signature, rmse, mae, r2

def build_train():
    # Read the wine-quality csv file from the URL
    # data = pd.read_csv("./tests/resources/data/winequality-red.csv", sep=",", header=True)
    csv_url = (
        "http://archive.ics.uci.edu/ml/machine-learning-databases/wine-quality/winequality-red.csv"
    )
    try:
        print("Downloading dataset...")
        data = pd.read_csv(csv_url, sep=";")
    except Exception as e:
        logger.exception(
            "Unable to download training & test CSV, check your internet connection. Error: %s", e
        )

    # Split the data into training and test sets. (0.75, 0.25) split.
    train, test = train_test_split(data)

    # The predicted column is "quality" which is a scalar from [3, 9]
    train_x = train.drop(["quality"], axis=1)
    test_x = test.drop(["quality"], axis=1)
    train_y = train[["quality"]]
    test_y = test[["quality"]]

    alpha = 0.5
    l1_ratio = 0.5

    return train_x, train_y, test_x, test_y, alpha, l1_ratio

def main(log_model_option: dict = {"flavor": "mlflow"}, run_inside_mlflow_context: bool = True):
    """ Train and log model

    :param log_model_option: Log model using options:
                             if {"flavor": "mlflow"}: log model using mlflow log_model method
                             else if {"flavor": "_log_model"} log model using shipped brain _log_model function
                             else ig {"flavor": "upload_run" | "upload_model", args...} log model using named function with args
                             NB: input_example and signature are not required
    :param run_inside_mlflow_context: if True run log method from mlflow run context,
                                      otherwise use shippedbrain.log_flavor outside without mlflow run context
    """
    warnings.filterwarnings("ignore")
    np.random.seed(46)
    

    print("MLflow Tracking URI:", mlflow.get_tracking_uri())

    train_x, train_y, test_x, test_y, alpha, l1_ratio = build_train()

    lr, predicted_qualities, signature, rmse, mae, r2 = train_and_eval(alpha,
                                                                       l1_ratio,
                                                                       train_x,
                                                                       train_y,
                                                                       test_x,
                                                                       test_y)

    log_model_option["signature"] = signature
    log_model_option["input_example"] = test_x.iloc[0:2]

    print(f"[INFO] RUN INSIDE MLFLOW RUN CONTEXT={run_inside_mlflow_context}")
    if run_inside_mlflow_context:
        with mlflow.start_run() as run:
            print("[INFO] Starting run with id:", run.info.run_id)

            print("[INFO]Elasticnet model (alpha=%f, l1_ratio=%f):" % (alpha, l1_ratio))
            print("[INFO]\tRMSE: %s" % rmse)
            print("[INFO]\tMAE: %s" % mae)
            print("[INFO]\tR2: %s" % r2)

            mlflow.log_param("alpha", alpha)
            mlflow.log_param("l1_ratio", l1_ratio)
            mlflow.log_metric("rmse", rmse)
            mlflow.log_metric("r2", r2)
            mlflow.log_metric("mae", mae)

            # Model registry does not work with file store
            print("[DEBUG] Log model option flavor:", log_model_option["flavor"])
            if log_model_option["flavor"] == "_log_flavor":
                _ = shippedbrain._log_flavor("sklearn", sk_model = lr, signature = signature, input_example = log_model_option["input_example"], artifact_path="model")
            # INTEGRATION
            elif log_model_option["flavor"] == "upload_model" or log_model_option["flavor"] == "upload_run":
                flavor = log_model_option["flavor"]
                log_model_option.pop("flavor")
                log_func = eval(f"shippedbrain.{flavor}")
                _ = log_func(**log_model_option)
            elif log_model_option["flavor"] == "mlflow":
                mlflow.sklearn.log_model(lr, "model", signature=signature, input_example=log_model_option["input_example"])

            print(f"[INFO] Model URI runs:/{run.info.run_id}/model\n")

        return run

    else:
        if log_model_option["flavor"] == "_log_flavor":
            run = shippedbrain._log_flavor("sklearn",
                                           sk_model=lr,
                                           signature=signature,
                                           input_example=log_model_option["input_example"],
                                           artifact_path="model")
        # INTEGRATION
        elif log_model_option["flavor"] == "upload_model" or log_model_option["flavor"] == "upload_run":
            flavor = log_model_option["flavor"]
            log_model_option.pop("flavor")
            log_func = eval(f"shippedbrain.{flavor}")
            run = log_func(**log_model_option)

        print(f"[INFO] Model URI runs:/{run.info.run_id}/model\n")

        return run



from setuptools import setup, find_packages

VERSION = '0.0.5'
DESCRIPTION = 'Client library to create serverless REST endpoints on shippedbrain.com'
LONG_DESCRIPTION = open("README.md", "r", encoding="utf-8").read()

install_requires = [
      "mlflow",
      "click"
]

extras_require = {
      "dev": ["pytest"]
}

# Setting up
setup(name="shippedbrain",
      version=VERSION,
      author="Shipped Brain",
      author_email="info@shippedbrain.com",
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type="text/markdown",
      packages=find_packages(),
      python_requires=">=3.6.0",
      install_requires=install_requires,
      extras_require=extras_require,
      keywords=['pretrained-models', 'shipped-brain', 'machine-learning', 'artificial-intelligence', 'deploy', 'serve', "serverless", "model-hub", "data-science", 'mlflow'],
      license="LICENSE",
      url="https://github.com/shippedbrain/shippedbrain",
      entry_points={
            "console_scripts": [
                  "shippedbrain=shippedbrain.cli:cli"
            ]
      },
      classifiers=[
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
      ])
